
cdef extern from "Python.h":
    #string manipulations
    object PyString_FromStringAndSize(char* str, int size)
    int PyString_Check(object o) 
    int PyString_Size(object)
    
    int PyObject_CheckReadBuffer(object o)
    int PyObject_AsReadBuffer(object obj, void **buffer, int *buffer_len)
    int PyObject_AsWriteBuffer(object obj, void **buffer, int *buffer_len)

    #refcounting
    void Py_INCREF(object o)
    void Py_DECREF(object o)
    
    #integer check
    int PyInt_Check(object o) 
    

ctypedef struct PyObject:
    unsigned int ob_refcnt

cdef extern from "string.h":
    int memcmp(void*, void*, int)
    int memcpy(void*, void*, int)
    int memset(void*, int, int)

cdef extern from "ctype.h":
    int isspace(int C)
    int isalpha(int C)
    int isupper(int C)

cdef extern from "stringview_helper.h":
    int stringview_fastsearch(char* s, int n, char* p, int m)
    int stringview_rfastsearch(char* s, int n, char* p, int m)
    int stringview_count(char* s, int n, char* p, int m)
    int stringview_findwhitespace(char* s, int n, int* start, int* length)
    int stringview_rfindwhitespace(char* s, int n, int* start, int* length)
    unsigned char whitespace[]
    int whitespace_length
    int stringview_strip(unsigned char* s, int n, unsigned char* p, int m, int dire)
    int result_table[]
    int stringview_compare(char* l, int ll, char* r, int rl, int op)
    int stringview_hash(unsigned char* p, int l)
    int stringview_check(unsigned char* p, int l, int which)
    int stringview_istitle(unsigned char* p, int l)
    void stringview_ul_translate(unsigned char* p, int l, int which, int incr)
    void stringview_swapcase(unsigned char* p, int l)
    int stringview_expandtabs_size(char* p, int size, int tabsize)
    void stringview_expandtabs(char* p, int size, char* q, int tabsize)
    

cdef int get_buffer(obj, char **data, int *length) except -1:
    cdef int result
    cdef void *vd
    result = PyObject_AsReadBuffer(obj, &vd, length)
    data[0] = <char *>vd
    return result

cdef int get_wbuffer(obj, char **data, int *length) except -1:
    cdef int result
    cdef void *vd
    result = PyObject_AsWriteBuffer(obj, &vd, length)
    data[0] = <char *>vd
    return result

cdef int _cmp(x, y):
    cdef char *xb, *yb
    cdef int xl, yl, r
    if get_buffer(x, &xb, &xl):
        raise TypeError("could not get buffer for argument 1")
    if get_buffer(y, &yb, &yl):
        raise TypeError("could not get buffer for argument 2")
    
    r = xl
    if yl < r:
        r = yl
    
    if r:
        r = memcmp(xb, yb, r)
    
    if r == 0:
        if xl == yl:
            return 0
        elif xl < yl:
            return -1
        return 1
    return r

cdef object new_string(int newsize, char** buffer):
    cdef int size
    st = PyString_FromStringAndSize(NULL, newsize)
    if get_buffer(st, buffer, &size):
        raise TypeError("could not get buffer of the output string")
    
    if size != newsize:
        MemoryError("requested size and actual size are different")
    
    return st

cdef int _fix_index(int value, int size):
    if value < 0:
        value = value + size
        if value < 0:
            value = 0
    elif value > size:
        value = size
    return value

cdef class stringview:
    cdef int length, offset, hash, writable #the current view length, offset, hash, and writable
    cdef char* buffer #the current pointer
    cdef int olength, ooffset #the original view length and offsets
    cdef orig         #the original object
    
    def __init__(self, st, int start=0, int stop=2147483647):
        cdef char* buffer
        cdef int size, case
        cdef stringview sta
        
        if PyObject_CheckReadBuffer(st) == 0:
            raise TypeError("could not get buffer")
        
        # three cases:
        #   stringview as input
        #   string or immutable buffer as input
        #   mutable object with possible resize: bytes,array,mmap
        
        if isinstance(st, stringview):
            sta = st
            _fix_sizes(sta)
            st = sta.orig
            start = _fix_index(start, sta.length) + sta.ooffset
            stop = _fix_index(stop, sta.length) + sta.ooffset
        
        self.writable = 0
        if PyString_Check(st) == 0:
            try:
                if get_wbuffer(st, &buffer, &size):
                    raise TypeError("could not get write buffer")
                self.writable = 1
            except TypeError:
                if get_buffer(st, &buffer, &size):
                    raise TypeError("could not get buffer")
        elif get_buffer(st, &buffer, &size):
            raise TypeError("could not get buffer")
        
        start = _fix_index(start, size)
        stop = _fix_index(stop, size)
        
        self.orig = st
        self.olength = self.length = stop-start
        self.buffer = buffer + start
        self.ooffset = self.offset = start
        self.hash = -1
    
#--------------------------- indexing and slicing ----------------------------
    def __getitem__(self, index):
        _fix_sizes(self)
        cdef int start, stop, step, size
        size = self.length
        if PyInt_Check(index):
            start = index
            if start < 0:
                start = start + size
                if start < 0:
                    raise IndexError("stringview index out of range")
            elif start >= size:
                raise IndexError("stringview index out of range")
            if size == 1:
                return self
            return stringview(self.orig, self.offset+start, self.offset+start+1)
        
        if index.step == 1:
            return self.__getslice__(index.start, index.stop)
        
        start, stop, step = index.indices(size)
        
        #handle slices where step != 1
        return NotImplemented
    
    def __getslice__(self, int start, int stop):
        _fix_sizes(self)
        cdef int size
        size = self.length
        
        start = _fix_index(start, size)
        stop = _fix_index(stop, size)
        
        if start == 0 and stop == size:
            return self
        
        return stringview(self.orig, self.offset+start, self.offset+stop)

#----------------------------------- find ------------------------------------

    def find(self, st):
        _fix_sizes(self)
        cdef char* buffer
        cdef int size
        if get_buffer(st, &buffer, &size):
            raise TypeError("could not get buffer")
        
        return stringview_fastsearch(self.buffer, self.length, buffer, size)
    
    def rfind(self, st):
        _fix_sizes(self)
        cdef char* buffer
        cdef int size
        if get_buffer(st, &buffer, &size):
            raise TypeError("could not get buffer")
        
        return stringview_rfastsearch(self.buffer, self.length, buffer, size)
    
    def index(self, st):
        cdef int result
        result = self.find(st)
        
        if result == -1:
            raise ValueError("substring not found")
        return result
    
    def rindex(self, st):
        cdef int result
        result = self.rfind(st)
        
        if result == -1:
            raise ValueError("substring not found")
        return result
    
    def __contains__(self, other):
        return self.find(other) != -1
    
    def startswith(self, other):
        return self[:len(other)] == other
    
    def endswith(self, other):
        return self[-len(other):] == other
    
    def count(self, other):
        _fix_sizes(self)
        cdef char* buffer
        cdef int size
        if get_buffer(other, &buffer, &size):
            raise TypeError("couldn't get buffer")
        
        if size == 0:
            return self.length + 1
        
        return stringview_count(self.buffer, self.length, buffer, size)
        
        
#------------------------------ transformations ------------------------------
    def expandtabs(self, int tabsize = 8):
        _fix_sizes(self)
        cdef int newsize
        cdef char* buffer
        buffer = self.buffer
        if tabsize < 0:
            tabsize = 0
        newsize = stringview_expandtabs_size(buffer, self.length, tabsize)
        if newsize == -1:
            return self
        
        st = new_string(newsize, &buffer)
        stringview_expandtabs(self.buffer, self.length, buffer, tabsize)
        return stringview(st)
        
    def capitalize(self):
        _fix_sizes(self)
        if self.length == 0 or isalpha(self.buffer[0]) == 0 or isupper(self.buffer[0]) == 1:
            return self
        
        cdef char* buffer
        st = new_string(self.length, &buffer)
        memcpy(buffer, self.buffer, self.length)
        buffer[0] = buffer[0] & 223
        return stringview(st)
    
    def lower(self):
        if self.islower():
            return self
        cdef char* buffer
        st = new_string(self.length, &buffer)
        memcpy(buffer, self.buffer, self.length)
        stringview_ul_translate(<unsigned char*>buffer, self.length, 5, 32)
        return stringview(st)
    
    def upper(self):
        if self.isupper():
            return self
        cdef char* buffer
        st = new_string(self.length, &buffer)
        memcpy(buffer, self.buffer, self.length)
        stringview_ul_translate(<unsigned char*>buffer, self.length, 3, -32)
        return stringview(st)
    
    def swapcase(self):
        _fix_sizes(self)
        cdef char* buffer
        st = new_string(self.length, &buffer)
        memcpy(buffer, self.buffer, self.length)
        stringview_swapcase(<unsigned char*>buffer, self.length)
        return stringview(st)
    
    def split(self, sep=None, int maxsplit=-1):
        if maxsplit < 0:
            maxsplit = 2147483647
        
        cdef stringview cur, se
        cdef int i, j, k
        
        _fix_sizes(self)
        lst = []
        cur = self
        if sep is None:
            #split on whitespace, but first strip leading and tailing whitespace
            cur = cur.strip()
            
            while maxsplit > 0:
                i = stringview_findwhitespace(cur.buffer, cur.length, &j, &k)
                if not i:
                    break
                lst.append(cur[:j])
                cur = cur[j+k:]
                maxsplit = maxsplit - 1
        else:
            se = stringview(sep)
            j = se.length
            
            while maxsplit > 0:
                i = stringview_fastsearch(cur.buffer, cur.length, se.buffer, j)
                if i == -1:
                    break
                lst.append(cur[:i])
                cur = cur[i+j:]
                maxsplit = maxsplit - 1
        lst.append(cur)
        return lst
    
    def rsplit(self, sep=None, int maxsplit=-1):
        if maxsplit < 0:
            maxsplit = 2147483647
        
        cdef stringview cur, se
        cdef int i, j, k
        
        _fix_sizes(self)
        lst = []
        cur = self
        if sep is None:
            #split on whitespace, but first strip leading and tailing whitespace
            cur = cur.strip()
            
            while maxsplit > 0:
                i = stringview_rfindwhitespace(cur.buffer, cur.length, &j, &k)
                if not i:
                    break
                lst.append(cur[j+k:])
                cur = cur[:j]
                maxsplit = maxsplit - 1
        else:
            se = stringview(sep)
            j = se.length
        
            while maxsplit > 0:
                i = stringview_rfastsearch(cur.buffer, cur.length, se.buffer, j)
                if i == -1:
                    break
                lst.append(cur[i+j:])
                cur = cur[:i]
                maxsplit = maxsplit - 1
        lst.append(cur)
        lst.reverse()
        return lst
        
    
    def replace(self, _old, _new, int count=-1):
        cdef stringview cur, new, old
        cdef int lo, p
        if count < 0:
            count = 2147483647
        
        #this is a fairly lazy implementation of replace...
        new = stringview(_new)
        old = stringview(_old)
        
        if old:
            return new.join(self.split(old, count))
        
        return NotImplemented
    
#--------------------------------- partition ---------------------------------

    cdef _partition(self, st, int dire):
        cdef int posn, start, stop, size
        
        if dire == 0:
            posn = self.find(st)
        else:
            posn = self.rfind(st)
        
        size = len(st)
        start = self.offset
        stop = self.offset+self.length
        orig = self.orig
        if posn == -1:
            x = stringview(orig, stop, stop)
            return self, x, x
        
        x = stringview(orig, start, start+posn)
        y = stringview(orig, start+posn, start+posn+size)
        z = stringview(orig, start+posn+size, stop)
        
        return x, y, z
    
    def partition(self, st):
        return self._partition(st, 0)
    
    def rpartition(self, st):
        return self._partition(st, 1)
    
#----------------------------------- strip -----------------------------------

    cdef stringview _strip(self, st, int dire):
        _fix_sizes(self)
        if self.length == 0:
            return self
        
        cdef int start, length, s, l
        cdef char* buffer
        
        cdef char* tostrip
        cdef int size
        if st is not None and get_buffer(st, &tostrip, &size):
            raise TypeError("could not get buffer")
        
        s = 0
        l = 0
        
        buffer = self.buffer
        start = 0
        length = self.length
        
        if dire != 1:
            l = 0
            if st is None:
                if isspace(buffer[0]):
                    stringview_findwhitespace(buffer, length, &s, &l)
            else:
                l = stringview_strip(<unsigned char*>buffer, length, <unsigned char*>tostrip, size, 1)
            start = start + l
        
        if dire != 0:
            l = 0
            if st is None:
                if isspace(buffer[length-1]):
                    stringview_rfindwhitespace(buffer+start, length-start, &s, &l)
            else:
                l = stringview_strip(<unsigned char*>(buffer+start), length-start, <unsigned char*>tostrip, size, -1)
            length = length - l
        
        if start == 0 and length == self.length:
            return self
        
        return stringview(self.orig, self.offset+start, self.offset+length)
    
    def lstrip(self, st=None):
        return self._strip(st, 0)
    
    def rstrip(self, st=None):
        return self._strip(st, 1)
    
    def strip(self, st=None):
        return self._strip(st, 2)

#-------------------------------- view adding --------------------------------
    
    def __add__(self, other):
        if isinstance(self, stringview):
            if not isinstance(other, stringview):
                other = stringview(other)
        else:
            self = stringview(self)
        cdef stringview x, y
        x = self
        y = other

        #we know both x and y are stringviews
        _fix_sizes(x)
        _fix_sizes(y)
        
        if x.length == 0:
            return y
        elif y.length == 0:
            return x
        elif x.orig is y.orig and x.offset+x.length == y.offset:
            return stringview(x.orig, x.offset, x.offset+x.length+y.length)
        
        #x and y either aren't the same object, or they aren't contiguous
        
        cdef char* buffer
        st = new_string(newsize, &buffer)
        
        memcpy(<void*>buffer, <void*>x.buffer, x.length)
        memcpy(<void*>(buffer+x.length), <void*>y.buffer, y.length)
        
        return stringview(st)
    
    def __mul__(x, y):
        if PyInt_Check(x) or PyInt_Check(y):
            return _st_repeat(x,y)
        return NotImplemented
        
    def join(self, seq):
        cdef char *buffer, *lastptr
        cdef int seen, finalsize, size, i
        _fix_sizes(self)
        if not len(seq):
            return stringview(self.orig, self.ooffset, self.ooffset)
        seq = list(seq)
        finalsize = 0
        seen = 0
        fastpath = 1
        # find the required length of the output
        for obj in seq:
            if isinstance(obj, stringview):
                _fix_sizes(obj)
            
            if get_buffer(obj, &buffer, &size):
                raise ValueError("value in list does not support the buffer protocol")
            
            if seen:
                finalsize = finalsize + self.length
                if fastpath:
                    if self.length and self.buffer != lastptr:
                        fastpath = 0
                        print "fastpath died in separator!"
                    else:
                        lastptr = lastptr + self.length
                        print "fastpath continued in separator!"
            else:
                lastptr = buffer
            
            finalsize = finalsize + size
            seen = 1
            if fastpath:
                if size and self.buffer != lastptr:
                    fastpath = 0
                    print "fastpath died"
                else:
                    lastptr = lastptr + size
                    print "fastpath continued"
            
        
        if fastpath:
            return stringview(self.orig, self.offset, self.offset+finalsize)
        
        #construct the string
        cdef char* output
        st = new_string(finalsize, &output)
        
        #copy the data
        seen = 0
        for obj in seq:
            if get_buffer(obj, &buffer, &size):
                raise ValueError("value in list does not support the buffer protocol")
            if seen:
                memcpy(output, self.buffer, self.length)
                output = output + self.length
            else:
                seen = 1
            memcpy(output, buffer, size)
            output = output + size
        
        return stringview(st)

#----------------------------------- tests -----------------------------------
    
    cdef int _ischeck(self, int which):
        _fix_sizes(self)
        return stringview_check(<unsigned char*>self.buffer, self.length, which)
    
    def isalnum(self):
        return self._ischeck(0)
    
    def isalpha(self):
        return self._ischeck(1)
    
    def isdigit(self):
        return self._ischeck(2)
    
    def islower(self):
        return self._ischeck(3)
    
    def isspace(self):
        return self._ischeck(4)

    def isupper(self):
        return self._ischeck(5)
    
    def istitle(self):
        _fix_sizes(self)
        return stringview_istitle(<unsigned char*>self.buffer, self.length)

#----------------------------- buffer interface ------------------------------

    cdef int __getsegcount__(self, int* p):
        return 1
    
    cdef int __getreadbuffer__(self, int i, void **p):
        _fix_sizes(self)
        if i != 0:
            raise ValueError("view only supports 1 segment")
        
        p[0] = <void*>self.buffer
        return self.length
    
    cdef int __getwritebuffer__(self, int i, void **p):
        if not self.writable:
            raise TypeError("buffer is not writable")
        return self.__getreadbuffer__(i, p)
    
    cdef int __getcharbuffer__(self, int i, void **p):
        return self.__getreadbuffer__(i, p)

#----------------------------------- misc ------------------------------------

    def __len__(self):
        _fix_sizes(self)
        return self.length
    
    def __nonzero__(self):
        _fix_sizes(self)
        return self.length > 0
    
    def __iter__(self):
        return stringview_iter(self)
    
    def __str__(self):
        _fix_sizes(self)
        if PyString_Check(self.orig) and self.offset == 0 and self.length == PyString_Size(self.orig):
            return self.orig
        return PyString_FromStringAndSize(self.buffer, self.length)
    
    def __repr__(self):
        return "<stringview at %08x of %08x>"%(id(self), id(self.orig))
    
    ## def __cmp__(self, st):
        ## #technically not necessary
        ## return _cmp(self, st)
    
    def __richcmp__(self, st, int op):
        #no need to _fix_sizes(), it is done in the get_buffer() calls
        cdef int r
        r = _cmp(self, st)
        op = _fix_index(op, 5)
        return bool(result_table[op+(6*(r+1))])
    
    cdef int __hash__(self):
        if self.writable == 1:
            raise TypeError("cannot hash mutable object")
        #_fix_sizes(self) #not necessary because self shouldn't resize
        if self.hash == -1:
            self.hash = stringview_hash(<unsigned char*>self.buffer, self.length)
        return self.hash

cdef class stringview_iter:
    cdef stringview view
    cdef int current
    def __init__(self, stringview view):
        self.view = view
        self.current = 0
    def __next__(self):
        _fix_sizes(self.view)
        if self.current >= self.view.length:
            raise StopIteration
        cdef stringview ret
        ret = self.view[self.current]
        self.current = self.current + 1
        return ret

cdef void _fix_sizes(stringview self):
    if self.writable == 0:
        return
    
    cdef char* buffer
    cdef int size
    if get_buffer(self.orig, &buffer, &size):
        raise TypeError("underlying object changed types?!?")
    
    if self.ooffset > size:
        self.offset = size
        self.length = 0
    if self.ooffset + self.olength > size:
        self.length = size - self.ooffset
    else:
        self.length = self.olength
    
    self.buffer = buffer+self.offset
    self.bsize = size

cdef stringview _st_repeat(x, y):
    cdef int count, newsize, i, size
    cdef stringview item
    if PyInt_Check(x):
        count = x
        item = y
    else:
        count = y
        item = x
    
    _fix_sizes(item)
    
    if count < 0:
        raise ValueError("Can't repeat a view less than 0 times")
    
    size = item.length
    
    newsize = count * size
    if newsize and count and ((newsize / count) != size):
        raise OverflowError("Repeated view is too long")
    
    if newsize == 0:
        return stringview(item.orig, item.offset, item.offset)
    
    cdef char* buffer
    st = new_string(newsize, &buffer)
    
    if size == 1:
        memset(<void*>buffer, <unsigned int>item.buffer[0], count)
    else:
        for i from 0 <= i < count:
            memcpy(<void*>buffer, item.buffer, size)
            buffer = buffer + size
    
    return stringview(st)




## * __add__
## * __contains__
## * __eq__             via __richcmp__
## * __ge__             via __richcmp__
## * __getitem__        don't have support for slices != 1 at the moment
## * __getslice__
## * __gt__             via __richcmp__
## * __hash__
## * __init__
## * __le__             via __richcmp__
## * __len__
## * __lt__             via __richcmp__
## - __mod__
## * __mul__
## * __ne__             via __richcmp__
## ? __new__            probably not necessary
## - __reduce__
## - __reduce_ex__
## * __repr__
## - __rmod__
## * __rmul__           __rmul__ is handled in __mul__
## * __setattr__
## * __str__
## * capitalize
## - center
## * count
## - decode
## - encode
## * endswith
## * expandtabs
## * find
## * index
## * isalnum
## * isalpha
## * isdigit
## * islower
## * isspace
## * istitle
## * isupper
## * join
## - ljust
## * lower
## * lstrip
## * partition
## * replace            don't have for old='' yet
## * rfind
## * rindex
## - rjust
## * rpartition
## * rsplit
## * rstrip
## * split
## - splitlines
## * startswith
## * strip
## * swapcase
## - title
## - translate
## * upper
## - zfill

